use std::sync::{
    atomic::{AtomicBool, Ordering},
    Arc,
};

use futures::stream::StreamExt;
use log::{debug, error, info, log_enabled, Level};
use netlink_packet_core::{NetlinkMessage, NetlinkPayload};
use netlink_packet_route::RouteNetlinkMessage as RtnlMessage;
use netlink_sys::{AsyncSocket, SocketAddr};
use rtnetlink::new_connection;

use dynip_cloudflare::{utils, CloudflareClient, MAX_ERORS_IN_ROW_DEFAULT};
use scopeguard::defer;
use tokio::{signal, sync::Notify};

const RTNLGRP_LINK: u32 = 1;
const RTNLGRP_IPV4_IFADDR: u32 = 5;

const fn nl_mgrp(group: u32) -> u32 {
    if group > 31 {
        panic!("use netlink_sys::Socket::add_membership() for this group");
    }
    if group == 0 {
        0
    } else {
        1 << (group - 1)
    }
}

#[tokio::main]
async fn main() {
    dynip_cloudflare::init_logger();
    defer! {
        log::logger().flush();
    }
    let should_exit = Arc::new(AtomicBool::new(false));
    let notify = Arc::new(Notify::new());

    let should_exit_clone = should_exit.clone();
    let notify_clone = notify.clone();

    tokio::spawn(async move {
        signal::ctrl_c()
            .await
            .expect("Failed to install CTRL+C signal handler");
        should_exit_clone.store(true, Ordering::SeqCst);
        notify_clone.notify_one();
    });

    let config = if let Some(aux) = utils::get_config().await {
        aux
    } else {
        return;
    };

    let mut cloudflare =
        match CloudflareClient::new(config.api_key, config.zone_id, config.domains).await {
            Ok(cloudflare) => cloudflare,
            Err(e) => {
                error!("Failed to initialize cloudflare client: {:?}", &e);
                return;
            }
        };

    let (mut conn, mut _handle, mut messages) = new_connection().unwrap();
    let groups = nl_mgrp(RTNLGRP_LINK) | nl_mgrp(RTNLGRP_IPV4_IFADDR);

    let addr = SocketAddr::new(0, groups);

    if let Err(e) = conn.socket_mut().socket_mut().bind(&addr) {
        error!("Failed to bind to socket: {:?}", &e);
        return;
    }

    tokio::spawn(conn);
    info!("Listening for IPv4 address changes and interface connect/disconnect events...");

    let mut errs_counter: usize = 0;
    let errs_max = config.max_errors_in_row.unwrap_or(MAX_ERORS_IN_ROW_DEFAULT);

    while !should_exit.load(Ordering::SeqCst) {
        tokio::select! {
            _ = notify.notified() => break,
            message = messages.next() => {
                if let Some((message, _)) = message {
                    handle_message(message, &mut cloudflare, &mut errs_counter, errs_max).await;
                }
            }
        }
    }
}

async fn handle_message(
    message: NetlinkMessage<RtnlMessage>,
    cloudflare: &mut CloudflareClient,
    errs_counter: &mut usize,
    errs_max: usize,
) {
    match message.payload {
        NetlinkPayload::InnerMessage(RtnlMessage::NewAddress(msg)) => {
            if log_enabled!(Level::Debug) {
                debug!("New IPv4 address message: {:?}", msg);
            } else {
                info!("New IPv4 address");
                if let Err(e) = cloudflare.check().await {
                    *errs_counter += 1;
                    error!(
                        "Failed to check cloudflare ({}/{}): {:?}",
                        errs_counter, errs_max, &e
                    );
                    if *errs_counter >= errs_max {
                        return;
                    }
                }
            }
        }
        NetlinkPayload::InnerMessage(RtnlMessage::DelAddress(msg)) => {
            if log_enabled!(Level::Debug) {
                debug!("Deleted IPv4 address message: {:?}", msg);
            } else {
                info!("Deleted IPv4 address");
            }
        }
        NetlinkPayload::InnerMessage(RtnlMessage::NewLink(link)) => {
            if log_enabled!(Level::Debug) {
                debug!("New link message (interface connected): {:?}", link);
            } else {
                info!("New link (interface connected)");
                if let Err(e) = cloudflare.check().await {
                    *errs_counter += 1;
                    error!(
                        "Failed to check cloudflare ({}/{}): {:?}",
                        errs_counter, errs_max, &e
                    );
                    if *errs_counter >= errs_max {
                        return;
                    }
                }
            }
        }
        NetlinkPayload::InnerMessage(RtnlMessage::DelLink(link)) => {
            if log_enabled!(Level::Debug) {
                debug!("Deleted link message (interface disconnected): {:?}", link);
            } else {
                info!("Deleted link (interface disconnected)");
            }
        }
        _ => {
            if log_enabled!(Level::Debug) {
                debug!("Unhandled message payload: {:?}", message.payload);
            }
        }
    }
}
